import typer
import importlib
import sys
from ..mqtt import (
    start_listener as start_listener_,
    run_pipeline_on_message as run_pipeline_on_message_,
)
from .utils import parse_dict_or_list_param

app = typer.Typer(help="MQTT management commands")


@app.command()
def start_listener(
    on_message: str,
    topic: str,
    base_dir: str,
    host: str = "localhost",
    port: int = 1883,
    username: str | None = None,
    password: str | None = None,
):
    """Start an MQTT client to listen to messages on a topic

    The connection to the MQTT broker is established using the provided configuration o a
    MQTT event broker defined in the project configuration file `conf/project.yml`.
    If not configuration is found, you have to provide the connection parameters,
    such as `host`, `port`, `username`, and `password`.

    The `on_message` module should contain a function `on_message` that will be called
    with the message payload as argument.

    Args:
        on_message: Name of the module containing the on_message function
        topic: MQTT topic to listen to
        base_dir: Base directory for the module
        host: MQTT broker host
        port: MQTT broker port
        username: MQTT broker username
        password: MQTT broker password

    Examples:
        $ flowerpower mqtt start_listener --on-message my_module --topic my_topic --base-dir /path/to/module
    """
    sys.path.append(base_dir)
    on_message_module = importlib.import_module(on_message)
    start_listener_(
        on_message=on_message_module.on_message,
        topic=topic,
        base_dir=base_dir,
        host=host,
        port=port,
        username=username,
        password=password,
        background=False,
    )


@app.command()
def run_pipeline_on_message(
    name: str,
    topic: str | None = None,
    executor: str | None = None,
    base_dir: str | None = None,
    inputs: str | None = None,
    final_vars: str | None = None,
    config: str | None = None,
    with_tracker: bool = False,
    with_opentelemetry: bool = False,
    with_progressbar: bool = False,
    storage_options: str | None = None,
    as_job: bool = False,
    host: str | None = None,
    port: int | None = None,
    username: str | None = None,
    password: str | None = None,
):
    """Run a pipeline on a message

    Args:
        name: Name of the pipeline
        topic: MQTT topic to listen to
        executor: Name of the executor
        base_dir: Base directory for the pipeline
        inputs: Inputs as JSON or key=value pairs or dict string
        final_vars: Final variables as JSON or list
        config: Config for the hamilton pipeline executor
        with_tracker: Enable tracking with hamilton ui
        with_opentelemetry: Enable OpenTelemetry tracing
        with_progressbar: Enable progress bar
        storage_options: Storage options as JSON, dict string or key=value pairs
        as_job: Run as a job in the scheduler
        host: MQTT broker host
        port: MQTT broker port
        username: MQTT broker username
        password: MQTT broker password


    """

    parsed_inputs = parse_dict_or_list_param(inputs, "dict")
    parsed_config = parse_dict_or_list_param(config, "dict")
    parsed_final_vars = parse_dict_or_list_param(final_vars, "list")
    parsed_storage_options = parse_dict_or_list_param(storage_options, "dict")

    run_pipeline_on_message_(
        name=name,
        topic=topic,
        executor=executor,
        base_dir=base_dir,
        inputs=parsed_inputs,
        final_vars=parsed_final_vars,
        config=parsed_config,
        with_tracker=with_tracker,
        with_opentelemetry=with_opentelemetry,
        with_progressbar=with_progressbar,
        storage_options=parsed_storage_options,
        as_job=as_job,
        host=host,
        port=port,
        username=username,
        password=password,
    )



from textual.app import App
from textual.layouts.grid import GridLayout
from textual.reactive import Reactive
from textual.scroll_view import ScrollView
from textual.widgets import Footer, Header, Static

from flowerpower.pipeline import PipelineManager


class PipelineList(Static):
    def __init__(self, pipelines):
        super().__init__()
        self.pipelines = pipelines

    def render(self):
        return "\n".join(f"- {pipeline.name}" for pipeline in self.pipelines)


class FlowerPowerTUI(App):
    pipelines = Reactive([])

    async def on_load(self):
        await self.bind("q", "quit")  # , "Beenden")
        await self.bind("a", "add_pipeline")  # , "Pipeline hinzufügen")
        await self.bind("d", "delete_pipeline")  # , "Pipeline löschen")

    async def on_mount(self):
        self.pipeline_manager = PipelineManager()
        self.refresh_pipelines()

        grid = await self.view.dock_grid(edge="top")

        grid.add_column(fraction=1, name="left", min_size=30)
        grid.add_column(fraction=3, name="right")

        grid.add_row(fraction=1, name="row")

        grid.add_areas(
            area_left="left,row",
            area_right="right,row",
        )

        header = Header()
        footer = Footer()

        self.pipeline_list = PipelineList(self.pipelines)
        self.details_view = ScrollView()

        await self.view.dock(header, edge="top")
        await self.view.dock(footer, edge="bottom")
        grid.place(area_left=self.pipeline_list, area_right=self.details_view)

        await self.view.dock(grid)

    def refresh_pipelines(self):
        self.pipelines = self.pipeline_manager.find_pipelines()

    async def action_add_pipeline(self):
        # Logik zum Hinzufügen einer neuen Pipeline
        pass

    async def action_delete_pipeline(self):
        # Logik zum Löschen einer Pipeline
        pass

    async def on_key(self, event):
        if event.key == "up":
            # Navigiere nach oben in der Pipeline-Liste
            pass
        elif event.key == "down":
            # Navigiere nach unten in der Pipeline-Liste
            pass
        elif event.key == "enter":
            # Zeige Details zur ausgewählten Pipeline an
            pass


if __name__ == "__main__":
    FlowerPowerTUI().run()



import datetime as dt
import importlib
import posixpath
import uuid
from typing import Generator

import orjson
import pandas as pd
import pyarrow as pa
import pyarrow.dataset as pds
import pyarrow.parquet as pq
from fsspec import AbstractFileSystem

from ..utils.misc import (
    _dict_to_dataframe,
    convert_large_types_to_standard,
    run_parallel,
    to_pyarrow_table,
)
from ..utils.polars import pl

if importlib.util.find_spec("duckdb") is not None:
    import duckdb
else:
    duckdb = None

if importlib.util.find_spec("pydala") is not None:
    from pydala.dataset import ParquetDataset
else:
    ParquetDataset = None


def path_to_glob(path: str, format: str | None = None) -> str:
    path = path.rstrip("/")
    if format is None:
        if ".json" in path:
            format = "json"
        elif ".csv" in path:
            format = "csv"
        elif ".parquet" in path:
            format = "parquet"

    if format in path:
        return path
    else:
        if path.endswith("**"):
            return posixpath.join(path, f"*.{format}")
        elif path.endswith("*"):
            if path.endswith("*/*"):
                return path + f".{format}"
            return posixpath.join(path.rstrip("/*"), f"*.{format}")
        return posixpath.join(path, f"**/*.{format}")


def _read_json_file(
    path, self, include_file_path: bool = False, jsonlines: bool = False
) -> dict | list[dict]:
    with self.open(path) as f:
        if jsonlines:
            data = [orjson.loads(line) for line in f.readlines()]
        else:
            data = orjson.loads(f.read())
    if include_file_path:
        return {path: data}
    return data


def read_json_file(
    self, path: str, include_file_path: bool = False, jsonlines: bool = False
) -> dict | list[dict]:
    return _read_json_file(
        path=path, self=self, include_file_path=include_file_path, jsonlines=jsonlines
    )


def _read_json(
    self,
    path: str | list[str],
    include_file_path: bool = False,
    use_threads: bool = True,
    jsonlines: bool = False,
    as_dataframe: bool = True,
    concat: bool = True,
    verbose: bool = False,
    **kwargs,
) -> dict | list[dict] | pl.DataFrame | list[pl.DataFrame]:
    """
    Read a JSON file or a list of JSON files.

    Args:
        path: (str | list[str]) Path to the JSON file(s).
        include_file_path: (bool, optional) If True, return a dictionary with the file path as key.
            Defaults to False.
        use_threads: (bool, optional) If True, read files in parallel. Defaults to True.
        jsonlines: (bool, optional) If True, read JSON lines. Defaults to False.
        as_dataframe: (bool, optional) If True, return a DataFrame. Defaults to True.
        concat: (bool, optional) If True, concatenate the DataFrames. Defaults to True.
        verbose: (bool, optional) If True, print verbose output. Defaults to False.
        **kwargs: Additional keyword arguments.

    Returns:
        (dict | list[dict] | pl.DataFrame | list[pl.DataFrame]):
            Dictionary, list of dictionaries, DataFrame or list of DataFrames.
    """
    if isinstance(path, str):
        path = path_to_glob(path, format="json")
        path = self.glob(path)

    if isinstance(path, list):
        if use_threads:
            data = run_parallel(
                _read_json_file,
                path,
                self=self,
                include_file_path=include_file_path,
                jsonlines=jsonlines,
                n_jobs=-1,
                backend="threading",
                verbose=verbose,
                **kwargs,
            )
        data = [
            _read_json_file(
                path=p,
                self=self,
                include_file_path=include_file_path,
                jsonlines=jsonlines,
            )
            for p in path
        ]
    else:
        data = _read_json_file(
            path=path,
            self=self,
            include_file_path=include_file_path,
            jsonlines=jsonlines,
        )
    if as_dataframe:
        if not include_file_path:
            data = [pl.DataFrame(d) for d in data]
        else:
            data = [
                [
                    pl.DataFrame(_data[k]).with_columns(pl.lit(k).alias("file_path"))
                    for k in _data
                ][0]
                for _data in data
            ]
        if concat:
            return pl.concat(data, how="diagonal_relaxed")
    return data


def _read_json_batches(
    self,
    path: str | list[str],
    batch_size: int | None = None,
    include_file_path: bool = False,
    jsonlines: bool = False,
    as_dataframe: bool = True,
    concat: bool = True,
    use_threads: bool = True,
    verbose: bool = False,
    **kwargs,
) -> Generator[dict | list[dict] | pl.DataFrame | list[pl.DataFrame], None, None]:
    """
    Read JSON files in batches with optional parallel processing within batches.

    Args:
        path: (str | list[str]) Path to the JSON file(s).
        batch_size: (int | None) Number of files to process in each batch. Defaults to None.
        include_file_path: (bool) If True, return with file path as key.
        jsonlines: (bool) If True, read JSON lines. Defaults to False.
        as_dataframe: (bool) If True, return DataFrame. Defaults to True.
        concat: (bool) If True, concatenate batch DataFrames. Defaults to True.
        use_threads: (bool) If True, use parallel processing within batches.
        verbose: (bool) If True, print verbose output.
        **kwargs: Additional keyword arguments.

    Yields:
        Data from num_batches files as dict/DataFrame based on parameters.
    """
    # Handle path resolution
    if isinstance(path, str):
        path = path_to_glob(path, format="json")
        path = self.glob(path)

    # Process files in batches
    for i in range(0, len(path), batch_size):
        batch_paths = path[i : i + batch_size]

        # Read batch with optional parallelization
        if use_threads and len(batch_paths) > 1:
            batch_data = run_parallel(
                _read_json_file,
                batch_paths,
                self=self,
                include_file_path=include_file_path,
                jsonlines=jsonlines,
                n_jobs=-1,
                backend="threading",
                verbose=verbose,
                **kwargs,
            )
        else:
            batch_data = [
                _read_json_file(
                    path=p,
                    self=self,
                    include_file_path=include_file_path,
                    jsonlines=jsonlines,
                )
                for p in batch_paths
            ]

        if as_dataframe:
            if not include_file_path:
                batch_dfs = [pl.DataFrame(d) for d in batch_data]
            else:
                batch_dfs = [
                    pl.DataFrame(list(d.values())[0]).with_columns(
                        pl.lit(list(d.keys())[0]).alias("file_path")
                    )
                    for d in batch_data
                ]

            if concat and len(batch_dfs) > 1:
                yield pl.concat(batch_dfs, how="diagonal_relaxed")
            else:
                yield batch_dfs
        else:
            yield batch_data


def read_json(
    self,
    path: str | list[str],
    batch_size: int | None = None,
    include_file_path: bool = False,
    jsonlines: bool = False,
    as_dataframe: bool = True,
    concat: bool = True,
    use_threads: bool = True,
    verbose: bool = False,
    **kwargs,
) -> (
    dict
    | list[dict]
    | pl.DataFrame
    | list[pl.DataFrame]
    | Generator[dict | list[dict] | pl.DataFrame | list[pl.DataFrame], None, None]
):
    """
    Read a JSON file or a list of JSON files. Optionally read in batches,
    returning a generator that sequentially yields data for specified number of files.

    Args:
        path: (str | list[str]) Path to the JSON file(s).
        batch_size: (int | None) Number of files to process in each batch. Defaults to None.
        include_file_path: (bool) If True, return with file path as key.
        jsonlines: (bool) If True, read JSON lines. Defaults to False.
        as_dataframe: (bool) If True, return DataFrame. Defaults to True.
        concat: (bool) If True, concatenate the DataFrames. Defaults to True.
        use_threads: (bool) If True, use parallel processing within batches. Defaults to True.
        verbose: (bool) If True, print verbose output. Defaults to False.
        **kwargs: Additional keyword arguments.

    Returns:
        (dict | list[dict] | pl.DataFrame | list[pl.DataFrame]:
            Dictionary, list of dictionaries, DataFrame, list of DataFrames
    Yields:
        (dict | list[dict] | pl.DataFrame | list[pl.DataFrame]:
            Dictionary, list of dictionaries, DataFrame, list of DataFrames
    """
    if batch_size is not None:
        return _read_json_batches(
            self=self,
            path=path,
            batch_size=batch_size,
            include_file_path=include_file_path,
            jsonlines=jsonlines,
            as_dataframe=as_dataframe,
            concat=concat,
            use_threads=use_threads,
            verbose=verbose,
            **kwargs,
        )
    return _read_json(
        self=self,
        path=path,
        include_file_path=include_file_path,
        jsonlines=jsonlines,
        as_dataframe=as_dataframe,
        concat=concat,
        use_threads=use_threads,
        verbose=verbose,
        **kwargs,
    )


def _read_csv_file(
    path, self, include_file_path: bool = False, **kwargs
) -> pl.DataFrame:
    print(path)
    with self.open(path) as f:
        df = pl.read_csv(f, **kwargs)
    if include_file_path:
        return df.with_columns(pl.lit(path).alias("file_path"))
    return df


def read_csv_file(
    self, path: str, include_file_path: bool = False, **kwargs
) -> pl.DataFrame:
    return _read_csv_file(
        path=path, self=self, include_file_path=include_file_path, **kwargs
    )


def _read_csv(
    self,
    path: str | list[str],
    include_file_path: bool = False,
    use_threads: bool = True,
    concat: bool = True,
    verbose: bool = False,
    **kwargs,
) -> pl.DataFrame | list[pl.DataFrame]:
    """
    Read a CSV file or a list of CSV files into a polars DataFrame.

    Args:
        path: (str | list[str]) Path to the CSV file(s).
        include_file_path: (bool, optional) If True, return a DataFrame with a 'file_path' column.
            Defaults to False.
        use_threads: (bool, optional) If True, read files in parallel. Defaults to True.
        concat: (bool, optional) If True, concatenate the DataFrames. Defaults to True.
        verbose: (bool, optional) If True, print verbose output. Defaults to False.
        **kwargs: Additional keyword arguments.

    Returns:
        (pl.DataFrame | list[pl.DataFrame]): Polars DataFrame or list of DataFrames.
    """
    if isinstance(path, str):
        path = path_to_glob(path, format="csv")
        path = self.glob(path)

    if isinstance(path, list):
        if use_threads:
            dfs = run_parallel(
                _read_csv_file,
                path,
                self=self,
                include_file_path=include_file_path,
                n_jobs=-1,
                backend="threading",
                verbose=verbose,
                **kwargs,
            )
        dfs = [
            _read_csv_file(p, self=self, include_file_path=include_file_path, **kwargs)
            for p in path
        ]
    else:
        dfs = _read_csv_file(
            path, self=self, include_file_path=include_file_path, **kwargs
        )
    if concat:
        return pl.concat(dfs, how="diagonal_relaxed")
    return dfs


def _read_csv_batches(
    self,
    path: str | list[str],
    batch_size: int | None = None,
    include_file_path: bool = False,
    concat: bool = True,
    use_threads: bool = True,
    verbose: bool = False,
    **kwargs,
) -> Generator[pl.DataFrame, None, None]:
    """
    Read CSV files in batches with optional parallel processing within batches.

    Args:
        path: (str | list[str]) Path to the CSV file(s).
        batch_size: (int | None) Number of files to process in each batch. Defaults to None.
        include_file_path: (bool) If True, include file_path column.
        concat: (bool) If True, concatenate batch DataFrames.
        use_threads: (bool) If True, use parallel processing within batches.
        verbose: (bool) If True, print verbose output.
        **kwargs: Additional keyword arguments.

    Yields:
        pl.DataFrame: DataFrame containing data from num_batches files.
    """
    # Handle path resolution
    if isinstance(path, str):
        path = path_to_glob(path, format="csv")
        path = self.glob(path)

    # Ensure path is a list
    if isinstance(path, str):
        path = [path]

    # Process files in batches
    for i in range(0, len(path), batch_size):
        batch_paths = path[i : i + batch_size]

        # Read batch with optional parallelization
        if use_threads and len(batch_paths) > 1:
            batch_dfs = run_parallel(
                _read_csv_file,
                batch_paths,
                self=self,
                include_file_path=include_file_path,
                n_jobs=-1,
                backend="threading",
                verbose=verbose,
                **kwargs,
            )
        else:
            batch_dfs = [
                _read_csv_file(
                    p, self=self, include_file_path=include_file_path, **kwargs
                )
                for p in batch_paths
            ]

        if concat and len(batch_dfs) > 1:
            yield pl.concat(batch_dfs, how="diagonal_relaxed")
        else:
            yield batch_dfs


def read_csv(
    self,
    path: str | list[str],
    batch_size: int | None = None,
    include_file_path: bool = False,
    concat: bool = True,
    use_threads: bool = True,
    verbose: bool = False,
    **kwargs,
) -> (
    pl.DataFrame
    | list[pl.DataFrame]
    | Generator[pl.DataFrame | list[pl.DataFrame], None, None]
):
    """
    Read a CSV file or a list of CSV files. Optionally read in batches,
    returning a generator that sequentially yields data for specified number of files.

    Args:
        path: (str | list[str]) Path to the CSV file(s).
        batch_size: (int | None) Number of files to process in each batch. Defaults to None.
        include_file_path: (bool, optional) If True, include 'file_path' column.
        concat: (bool, optional) If True, concatenate the batch DataFrames. Defaults to True.
        use_threads: (bool, optional) If True, use parallel processing within batches. Defaults to True.
        verbose: (bool, optional) If True, print verbose output. Defaults to False.
        **kwargs: Additional keyword arguments.

    Returns:
        pl.DataFrame | list[pl.DataFrame]:
            DataFrame or list or DataFrames containing data from num_batches files.

    Yields:
        pl.DataFrame | list[pl.DataFrame]:
            DataFrame or list of DataFrames containing data from num_batches files.
    """
    if batch_size is not None:
        return _read_csv_batches(
            self=self,
            path=path,
            batch_size=batch_size,
            include_file_path=include_file_path,
            concat=concat,
            use_threads=use_threads,
            verbose=verbose,
            **kwargs,
        )
    return _read_csv(
        self=self,
        path=path,
        include_file_path=include_file_path,
        concat=concat,
        use_threads=use_threads,
        verbose=verbose,
        **kwargs,
    )


def _read_parquet_file(
    path, self, include_file_path: bool = False, **kwargs
) -> pa.Table:
    table = pq.read_table(path, filesystem=self, **kwargs)
    if include_file_path:
        return table.add_column(0, "file_path", pl.Series([path] * table.num_rows))
    return table


def read_parquet_file(
    self, path: str, include_file_path: bool = False, **kwargs
) -> pa.Table:
    return _read_parquet_file(
        path=path, self=self, include_file_path=include_file_path, **kwargs
    )


def _read_parquet(
    self,
    path,
    include_file_path: bool = False,
    use_threads: bool = True,
    concat: bool = True,
    verbose: bool = False,
    **kwargs,
) -> pa.Table | list[pa.Table]:
    """
    Read a Parquet file or a list of Parquet files into a pyarrow Table.

    Args:
        path: (str | list[str]) Path to the Parquet file(s).
        include_file_path: (bool, optional) If True, return a Table with a 'file_path' column.
            Defaults to False.
        use_threads: (bool, optional) If True, read files in parallel. Defaults to True.
        concat: (bool, optional) If True, concatenate the Tables. Defaults to True.
        **kwargs: Additional keyword arguments.

    Returns:
        (pa.Table | list[pa.Table]): Pyarrow Table or list of Pyarrow Tables.
    """
    if not include_file_path and concat:
        if isinstance(path, str):
            path = path.replace("**", "").replace("*.parquet", "")
        return pq.read_table(path, filesystem=self, **kwargs)
    else:
        if isinstance(path, str):
            path = path_to_glob(path, format="parquet")
            path = self.glob(path)

        if isinstance(path, list):
            if use_threads:
                table = run_parallel(
                    _read_parquet_file,
                    path,
                    self=self,
                    include_file_path=include_file_path,
                    n_jobs=-1,
                    backend="threading",
                    verbose=verbose,
                    **kwargs,
                )
            else:
                table = [
                    _read_parquet_file(
                        p, self=self, include_file_path=include_file_path, **kwargs
                    )
                    for p in path
                ]
        else:
            table = _read_parquet_file(
                path=path, self=self, include_file_path=include_file_path, **kwargs
            )
    if concat:
        return pa.concat_tables(table, promote_options="permissive")
    return table


def _read_parquet_batches(
    self,
    path: str | list[str],
    batch_size: int | None = None,
    include_file_path: bool = False,
    use_threads: bool = True,
    concat: bool = True,
    verbose: bool = False,
    **kwargs,
) -> Generator[pa.Table | list[pa.Table], None, None]:
    """
    Read Parquet files in batches, yielding PyArrow Tables.

    Args:
        path: (str | list[str]) Path to the Parquet file(s).
        batch_size: (int | None) Number of files to process in each batch. Defaults to None.
        include_file_path: (bool) If True, return Tables with 'file_path' column. Defaults to False.
        use_threads: (bool) If True, read files in parallel within batches. Defaults to True.
        concat: (bool) If True, concatenate Tables within each batch. Defaults to True.
        verbose: (bool) If True, print progress information. Defaults to False.
        **kwargs: Additional keyword arguments.

    Yields:
        pa.Table | list[pa.Table]: Table or list of Tables per batch.
    """
    # Fast path for simple cases
    if not include_file_path and concat and batch_size is None:
        if isinstance(path, str):
            path = path.replace("**", "").replace("*.parquet", "")
        yield pq.read_table(path, filesystem=self, **kwargs)
        return

    # Resolve path(s) to list
    if isinstance(path, str):
        path = path_to_glob(path, format="parquet")
        path = self.glob(path)

    if not isinstance(path, list):
        yield _read_parquet_file(
            path=path, self=self, include_file_path=include_file_path, **kwargs
        )
        return

    # Process in batches

    for i in range(0, len(path), batch_size):
        batch_paths = path[i : i + batch_size]
        if use_threads and len(batch_paths) > 1:
            batch_tables = run_parallel(
                _read_parquet_file,
                batch_paths,
                self=self,
                include_file_path=include_file_path,
                n_jobs=-1,
                backend="threading",
                verbose=verbose,
                **kwargs,
            )
        else:
            batch_tables = [
                _read_parquet_file(
                    p, self=self, include_file_path=include_file_path, **kwargs
                )
                for p in batch_paths
            ]

        if concat and batch_tables:
            yield pa.concat_tables(batch_tables, promote_options="permissive")
        else:
            yield batch_tables


def read_parquet(
    self,
    path: str | list[str],
    batch_size: int | None = None,
    include_file_path: bool = False,
    concat: bool = True,
    use_threads: bool = True,
    verbose: bool = False,
    **kwargs,
) -> pa.Table | list[pa.Table] | Generator[pa.Table | list[pa.Table], None, None]:
    """
    Read a Parquet file or a list of Parquet files. Optionally read in batches,
    returning a generator that sequentially yields data for specified number of files.

    Args:
        path: (str | list[str]) Path to the Parquet file(s).
        batch_size: (int | None) Number of files to process in each batch. Defaults to None.
        include_file_path: (bool, optional) If True, include 'file_path' column.
        concat: (bool, optional) If True, concatenate the batch Tables. Defaults to True.
        use_threads: (bool, optional) If True, use parallel processing within batches. Defaults to True.
        verbose: (bool, optional) If True, print verbose output. Defaults to False.
        **kwargs: Additional keyword arguments.

    Returns:
        pa.Table | list[pa.Table]:
            PyArrow Table or list of PyArrow Tables containing data from num_batches files.

    Yields:
        pa.Table | list[pa.Table]: PyArrow Table or list of PyArrow Tables containing data from num_batches files.
    """
    if batch_size is not None:
        return _read_parquet_batches(
            self=self,
            path=path,
            batch_size=batch_size,
            include_file_path=include_file_path,
            concat=concat,
            use_threads=use_threads,
            verbose=verbose,
            **kwargs,
        )
    return _read_parquet(
        self=self,
        path=path,
        include_file_path=include_file_path,
        use_threads=use_threads,
        concat=concat,
        verbose=verbose,
        **kwargs,
    )


def read_files(
    self,
    path: str | list[str],
    format: str,
    batch_size: int | None = None,
    include_file_path: bool = False,
    concat: bool = True,
    jsonlines: bool = False,
    use_threads: bool = True,
    verbose: bool = False,
    **kwargs,
) -> (
    pl.DataFrame
    | pa.Table
    | list[pl.DataFrame]
    | list[pa.Table]
    | Generator[
        pl.DataFrame | pa.Table | list[pa.Table] | list[pl.DataFrame], None, None
    ]
):
    """
    Read a file or a list of files of the given format.

    Args:
        path: (str | list[str]) Path to the file(s).
        format: (str) Format of the file.
        batch_size: (int | None) Number of files to process in each batch. Defaults to None.
        include_file_path: (bool, optional) If True, return a DataFrame with a 'file_path' column.
            Defaults to False.
        concat: (bool, optional) If True, concatenate the DataFrames. Defaults to True.
        jsonlines: (bool, optional) If True, read JSON lines. Defaults to False.
        use_threads: (bool, optional) If True, read files in parallel. Defaults to True.
        verbose: (bool, optional) If True, print verbose output. Defaults to False.
        **kwargs: Additional keyword arguments.

    Returns:
        (pl.DataFrame | pa.Table | list[pl.DataFrame] | list[pa.Table]):
            Polars DataFrame, Pyarrow Table or list of DataFrames, LazyFrames or Tables.

    Yields:
        (pl.DataFrame | pa.Table):
            Polars DataFrame, Pyarrow Table or list of DataFrames, LazyFrames or Tables.
    """
    if format == "json":
        if batch_size is not None:
            return read_json(
                self=self,
                path=path,
                batch_size=batch_size,
                include_file_path=include_file_path,
                jsonlines=jsonlines,
                concat=concat,
                use_threads=use_threads,
                verbose=verbose,
                **kwargs,
            )
        return read_json(
            self,
            path,
            include_file_path=include_file_path,
            jsonlines=jsonlines,
            concat=concat,
            use_threads=use_threads,
            verbose=verbose,
            **kwargs,
        )
    elif format == "csv":
        if batch_size is not None:
            return read_csv(
                self=self,
                path=path,
                batch_size=batch_size,
                include_file_path=include_file_path,
                concat=concat,
                use_threads=use_threads,
                verbose=verbose,
                **kwargs,
            )
        return read_csv(
            self,
            path,
            include_file_path=include_file_path,
            use_threads=use_threads,
            concat=concat,
            verbose=verbose,
            **kwargs,
        )
    elif format == "parquet":
        if batch_size is not None:
            return read_parquet(
                self=self,
                path=path,
                batch_size=batch_size,
                include_file_path=include_file_path,
                concat=concat,
                use_threads=use_threads,
                verbose=verbose,
                **kwargs,
            )
        return read_parquet(
            self,
            path,
            include_file_path=include_file_path,
            use_threads=use_threads,
            concat=concat,
            verbose=verbose,
            **kwargs,
        )


def pyarrow_dataset(
    self,
    path: str,
    format="parquet",
    schema: pa.Schema | None = None,
    partitioning: str | list[str] | pds.Partitioning = None,
    **kwargs,
) -> pds.Dataset:
    """
    Create a pyarrow dataset.

    Args:
        path: (str) Path to the dataset.
        format: (str, optional) Format of the dataset. Defaults to 'parquet'.
        schema: (pa.Schema, optional) Schema of the dataset. Defaults to None.
        partitioning: (str | list[str] | pds.Partitioning, optional) Partitioning of the dataset.
            Defaults to None.
        **kwargs: Additional keyword arguments.

    Returns:
        (pds.Dataset): Pyarrow dataset.
    """
    return pds.dataset(
        path,
        filesystem=self,
        partitioning=partitioning,
        schema=schema,
        format=format,
        **kwargs,
    )


def pyarrow_parquet_dataset(
    self,
    path: str,
    schema: pa.Schema | None = None,
    partitioning: str | list[str] | pds.Partitioning = None,
    **kwargs,
) -> pds.Dataset:
    """
    Create a pyarrow dataset from a parquet_metadata file.

    Args:
        path: (str) Path to the dataset.
        schema: (pa.Schema, optional) Schema of the dataset. Defaults to None.
        partitioning: (str | list[str] | pds.Partitioning, optional) Partitioning of the dataset.
            Defaults to None.
        **kwargs: Additional keyword arguments.

    Returns:
        (pds.Dataset): Pyarrow dataset.
    """
    if not self.is_file(path):
        path = posixpath.join(path, "_metadata")
    return pds.dataset(
        path,
        filesystem=self,
        partitioning=partitioning,
        schema=schema,
        **kwargs,
    )


def pydala_dataset(
    self,
    path: str,
    partitioning: str | list[str] | pds.Partitioning = None,
    **kwargs,
) -> ParquetDataset:  # type: ignore
    """
    Create a pydala dataset.

    Args:
        path: (str) Path to the dataset.
        partitioning: (str | list[str] | pds.Partitioning, optional) Partitioning of the dataset.
            Defaults to None.
        **kwargs: Additional keyword arguments.

    Returns:
        (ParquetDataset): Pydala dataset.
    """
    return ParquetDataset(
        path,
        filesystem=self,
        partitioning=partitioning,
        **kwargs,
    )


def write_parquet(
    self,
    data: pl.DataFrame | pl.LazyFrame | pa.Table | pd.DataFrame | dict | list[dict],
    path: str,
    schema: pa.Schema | None = None,
    **kwargs,
) -> pq.FileMetaData:
    """
    Write a DataFrame to a Parquet file.

    Args:
        data: (pl.DataFrame | pl.LazyFrame | pa.Table | pd.DataFrame) Data to write.
        path: (str) Path to write the data.
        schema: (pa.Schema, optional) Schema of the data. Defaults to None.
        **kwargs: Additional keyword arguments for `pq.write_table`.

    Returns:
        (pq.FileMetaData): Parquet file metadata.
    """
    data = to_pyarrow_table(data, concat=False, unique=False)

    if schema is not None:
        data = data.cast(schema)
    metadata = []
    pq.write_table(data, path, filesystem=self, metadata_collector=metadata, **kwargs)
    metadata = metadata[0]
    metadata.set_file_path(path)
    return metadata


def write_json(
    self,
    data: (
        dict | pl.DataFrame | pl.LazyFrame | pa.Table | pd.DataFrame | dict | list[dict]
    ),
    path: str,
    append: bool = False,
) -> None:
    """
    Write a dictionary, DataFrame or Table to a JSON file.

    Args:
        data: (dict | pl.DataFrame | pl.LazyFrame | pa.Table | pd.DataFrame) Data to write.
        path: (str) Path to write the data.
        append: (bool, optional) If True, append to the file. Defaults to False.

    Returns:
        None
    """
    if isinstance(data, pl.LazyFrame):
        data = data.collect()
    if isinstance(data, pl.DataFrame):
        data = data.to_arrow()
        data = data.cast(convert_large_types_to_standard(data.schema)).to_pydict()
    elif isinstance(data, pd.DataFrame):
        data = pa.Table.from_pandas(data, preserve_index=False).to_pydict()
    elif isinstance(data, pa.Table):
        data = data.to_pydict()
    if append:
        with self.open(path, "ab") as f:
            f.write(orjson.dumps(data))
            f.write(b"\n")
    else:
        with self.open(path, "wb") as f:
            f.write(orjson.dumps(data))


def write_csv(
    self,
    data: pl.DataFrame | pl.LazyFrame | pa.Table | pd.DataFrame | dict,
    path: str,
    **kwargs,
) -> None:
    """
    Write a DataFrame to a CSV file.

    Args:
        data: (pl.DataFrame | pl.LazyFrame | pa.Table | pd.DataFrame) Data to write.
        path: (str) Path to write the data.
        **kwargs: Additional keyword arguments for `pl.DataFrame.write_csv`.

    Returns:
        None
    """
    if isinstance(data, dict | list):
        data = _dict_to_dataframe(data)
    elif isinstance(data, pl.LazyFrame):
        data = data.collect()
    elif isinstance(data, pa.Table):
        data = pl.from_arrow(data)
    elif isinstance(data, pd.DataFrame):
        data = pl.from_pandas(data)

    with self.open(path, "w") as f:
        data.write_csv(f, **kwargs)


def write_file(
    self,
    data: pl.DataFrame | pl.LazyFrame | pa.Table | pd.DataFrame | dict | list[dict],
    path: str,
    format: str,
    **kwargs,
) -> None:
    """
    Write a DataFrame to a file in the given format.

    Args:
        data: (pl.DataFrame | pl.LazyFrame | pa.Table | pd.DataFrame) Data to write.
        path (str): Path to write the data.
        format (str): Format of the file.
        **kwargs: Additional keyword arguments.

    Returns:
        None
    """
    if format == "json":
        write_json(self, data, path, **kwargs)
    elif format == "csv":
        write_csv(self, data, path, **kwargs)
    elif format == "parquet":
        write_parquet(self, data, path, **kwargs)


def write_files(
    self,
    data: (
        pl.DataFrame
        | pl.LazyFrame
        | pa.Table
        | pa.RecordBatch
        | pa.RecordBatchReader
        | pd.DataFrame
        | dict
        | list[
            pl.DataFrame
            | pl.LazyFrame
            | pa.Table
            | pa.RecordBatch
            | pa.RecordBatchReader
            | pd.DataFrame
            | dict
        ]
    ),
    path: str | list[str],
    basename: str = None,
    format: str = None,
    concat: bool = True,
    unique: bool | list[str] | str = False,
    mode: str = "append",  # append, overwrite, delete_matching, error_if_exists
    use_threads: bool = True,
    verbose: bool = False,
    **kwargs,
) -> None:
    """Write a DataFrame or a list of DataFrames to a file or a list of files.

    Args:
        data: (pl.DataFrame | pl.LazyFrame | pa.Table | pd.DataFrame | dict | list[pl.DataFrame | pl.LazyFrame |
            pa.Table | pd.DataFrame | dict]) Data to write.
        path: (str | list[str]) Path to write the data.
        basename: (str, optional) Basename of the files. Defaults to None.
        format: (str, optional) Format of the data. Defaults to None.
        concat: (bool, optional) If True, concatenate the DataFrames. Defaults to True.
        unique: (bool | list[str] | str, optional) If True, remove duplicates. Defaults to False.
        mode: (str, optional) Write mode. Defaults to 'append'. Options: 'append', 'overwrite', 'delete_matching',
            'error_if_exists'.
        use_threads: (bool, optional) If True, use parallel processing. Defaults to True.
        verbose: (bool, optional) If True, print verbose output. Defaults to False.
        **kwargs: Additional keyword arguments.

    Returns:
        None

    Raises:
        FileExistsError: If file already exists and mode is 'error_if_exists'.
    """
    if not isinstance(data, list):
        data = [data]

    if concat:
        if isinstance(data[0], dict):
            data = _dict_to_dataframe(data)
        if isinstance(data[0], pl.LazyFrame):
            data = pl.concat([d.collect() for d in data], how="diagonal_relaxed")

        if isinstance(
            data[0], pa.Table | pa.RecordBatch | pa.RecordBatchReader | Generator
        ):
            data = pl.concat([pl.from_arrow(d) for d in data], how="diagonal_relaxed")
        elif isinstance(data[0], pd.DataFrame):
            data = pl.concat([pl.from_pandas(d) for d in data], how="diagonal_relaxed")

        if unique:
            data = data.unique(
                subset=None if not isinstance(unique, str | list) else unique,
                maintain_order=True,
            )

        data = [data]

    if format is None:
        format = (
            path[0].split(".")[-1]
            if isinstance(path, list) and "." in path[0]
            else path.split(".")[-1]
            if "." in path
            else "parquet"
        )

    def _write(d, p, basename, i):
        if f".{format}" not in p:
            if not basename:
                basename = f"data-{dt.datetime.now().strftime('%Y%m%d_%H%M%S%f')[:-3]}-{uuid.uuid4().hex[:16]}"
            p = f"{p}/{basename}-{i}.{format}"

        if mode == "delete_matching":
            write_file(self, d, p, format, **kwargs)
        elif mode == "overwrite":
            if self.exists(p):
                self.fs.rm(p, recursive=True)
            write_file(self, d, p, format, **kwargs)
        elif mode == "append":
            if not self.exists(p):
                write_file(self, d, p, format, **kwargs)
            else:
                p = p.replace(f".{format}", f"-{i}.{format}")
                write_file(self, d, p, format, **kwargs)
        elif mode == "error_if_exists":
            if self.exists(p):
                raise FileExistsError(f"File already exists: {p}")
            else:
                write_file(self, d, p, format, **kwargs)

    if mode == "overwrite":
        if isinstance(path, list):
            for p in path:
                # Remove existing files
                if self.exists(p):
                    self.rm(p, recursive=True)
        else:
            # Remove existing files
            if self.exists(path):
                self.rm(path, recursive=True)

    if use_threads:
        run_parallel(
            _write,
            d=data,
            p=path,
            basename=basename,
            i=list(range(len(data))),
            verbose=verbose,
        )
    else:
        for i, p in enumerate(path):
            _write(i, data, p, basename)


def write_pyarrow_dataset(
    self,
    data: (
        pl.DataFrame
        | pl.LazyFrame
        | pa.Table
        | pa.RecordBatch
        | pa.RecordBatchReader
        | pd.DataFrame
        | dict
        | list[
            pl.DataFrame
            | pl.LazyFrame
            | pa.Table
            | pa.RecordBatch
            | pa.RecordBatchReader
            | pd.DataFrame
            | dict
        ]
    ),
    path: str,
    basename: str | None = None,
    schema: pa.Schema | None = None,
    partition_by: str | list[str] | pds.Partitioning | None = None,
    partitioning_flavor: str = "hive",
    mode: str = "append",
    format: str | None = "parquet",
    compression: str = "zstd",
    max_rows_per_file: int | None = 2_500_000,
    row_group_size: int | None = 250_000,
    concat: bool = True,
    unique: bool | list[str] | str = False,
    **kwargs,
) -> list[pq.FileMetaData] | None:
    """
    Write a tabluar data to a PyArrow dataset.

    Args:
        data: (pl.DataFrame | pa.Table | pa.RecordBatch | pa.RecordBatchReader |
            pd.DataFrame | list[pl.DataFrame] | list[pa.Table] | list[pa.RecordBatch] |
            list[pa.RecordBatchReader] | list[pd.DataFrame]) Data to write.
        path: (str) Path to write the data.
        basename: (str, optional) Basename of the files. Defaults to None.
        schema: (pa.Schema, optional) Schema of the data. Defaults to None.
        partition_by: (str | list[str] | pds.Partitioning, optional) Partitioning of the data.
            Defaults to None.
        partitioning_flavor: (str, optional) Partitioning flavor. Defaults to 'hive'.
        mode: (str, optional) Write mode. Defaults to 'append'.
        format: (str, optional) Format of the data. Defaults to 'parquet'.
        compression: (str, optional) Compression algorithm. Defaults to 'zstd'.
        max_rows_per_file: (int, optional) Maximum number of rows per file. Defaults to 2_500_000.
        row_group_size: (int, optional) Row group size. Defaults to 250_000.
        concat: (bool, optional) If True, concatenate the DataFrames. Defaults to True.
        unique: (bool | list[str] | str, optional) If True, remove duplicates. Defaults to False.
        **kwargs: Additional keyword arguments for `pds.write_dataset`.

    Returns:
        (list[pq.FileMetaData] | None): List of Parquet file metadata or None.
    """
    data = to_pyarrow_table(data, concat=concat, unique=unique)

    if mode == "delete_matching":
        existing_data_behavior = "delete_matching"
    elif mode == "append":
        existing_data_behavior = "overwrite_or_ignore"
    elif mode == "overwrite":
        self.rm(path, recursive=True)
        existing_data_behavior = "overwrite_or_ignore"
    else:
        existing_data_behavior = mode

    if basename is None:
        basename_template = (
            "data-"
            f"{dt.datetime.now().strftime('%Y%m%d_%H%M%S%f')[:-3]}-{uuid.uuid4().hex[:16]}-{{i}}.parquet"
        )
    else:
        basename_template = f"{basename}-{{i}}.parquet"

    file_options = pds.ParquetFileFormat().make_write_options(compression=compression)

    create_dir: bool = (False,)

    if hasattr(self, "fs"):
        if "local" in self.fs.protocol:
            create_dir = True
    else:
        if "local" in self.protocol:
            create_dir = True

    if format == "parquet":
        metadata = []

        def file_visitor(written_file):
            file_metadata = written_file.metadata
            file_metadata.set_file_path(written_file.path)
            metadata.append(file_metadata)

    pds.write_dataset(
        data=data,
        base_dir=path,
        basename_template=basename_template,
        partitioning=partition_by,
        partitioning_flavor=partitioning_flavor,
        filesystem=self,
        existing_data_behavior=existing_data_behavior,
        min_rows_per_group=row_group_size,
        max_rows_per_group=row_group_size,
        max_rows_per_file=max_rows_per_file,
        schema=schema,
        format=format,
        create_dir=create_dir,
        file_options=file_options,
        file_visitor=file_visitor if format == "parquet" else None,
        **kwargs,
    )
    if format == "parquet":
        return metadata


def write_pydala_dataset(
    self,
    data: (
        pl.DataFrame
        | pl.LazyFrame
        | pa.Table
        | pa.RecordBatch
        | pa.RecordBatchReader
        | pd.DataFrame
        | dict
        | list[
            pl.DataFrame
            | pl.LazyFrame
            | pa.Table
            | pa.RecordBatch
            | pa.RecordBatchReader
            | pd.DataFrame
            | dict
        ]
    ),
    path: str,
    mode: str = "append",  # "delta", "overwrite"
    basename: str | None = None,
    partition_by: str | list[str] | None = None,
    partitioning_flavor: str = "hive",
    max_rows_per_file: int | None = 2_500_000,
    row_group_size: int | None = 250_000,
    compression: str = "zstd",
    concat: bool = True,
    sort_by: str | list[str] | list[tuple[str, str]] | None = None,
    unique: bool | str | list[str] = False,
    delta_subset: str | list[str] | None = None,
    update_metadata: bool = True,
    alter_schema: bool = False,
    timestamp_column: str | None = None,
    verbose: bool = False,
    **kwargs,
) -> None:
    """Write a tabular data to a Pydala dataset.

    Args:
        data: (pl.DataFrame | pa.Table | pa.RecordBatch | pa.RecordBatchReader |
            pd.DataFrame | list[pl.DataFrame] | list[pa.Table] | list[pa.RecordBatch] |
            list[pa.RecordBatchReader] | list[pd.DataFrame]) Data to write.
        path: (str) Path to write the data.
        mode: (str, optional) Write mode. Defaults to 'append'. Options: 'delta', 'overwrite'.
        basename: (str, optional) Basename of the files. Defaults to None.
        partition_by: (str | list[str], optional) Partitioning of the data. Defaults to None.
        partitioning_flavor: (str, optional) Partitioning flavor. Defaults to 'hive'.
        max_rows_per_file: (int, optional) Maximum number of rows per file. Defaults to 2_500_000.
        row_group_size: (int, optional) Row group size. Defaults to 250_000.
        compression: (str, optional) Compression algorithm. Defaults to 'zstd'.
        sort_by: (str | list[str] | list[tuple[str, str]], optional) Columns to sort by. Defaults to None.
        unique: (bool | str | list[str], optional) If True, ensure unique values. Defaults to False.
        delta_subset: (str | list[str], optional) Subset of columns to include in delta table. Defaults to None.
        update_metadata: (bool, optional) If True, update metadata. Defaults to True.
        alter_schema: (bool, optional) If True, alter schema. Defaults to False.
        timestamp_column: (str, optional) Timestamp column. Defaults to None.
        verbose: (bool, optional) If True, print verbose output. Defaults to False.
        **kwargs: Additional keyword arguments for `ParquetDataset.write_to_dataset`.

    Returns:
        None
    """
    data = to_pyarrow_table(data, concat=concat, unique=unique)

    ds = pydala_dataset(self=self, path=path, partitioning=partitioning_flavor)
    ds.write_to_dataset(
        data=data,
        mode=mode,
        basename=basename,
        partition_by=partition_by,
        max_rows_per_file=max_rows_per_file,
        row_group_size=row_group_size,
        compression=compression,
        sort_by=sort_by,
        unique=unique,
        delta_subset=delta_subset,
        update_metadata=update_metadata,
        alter_schema=alter_schema,
        timestamp_column=timestamp_column,
        verbose=verbose,
        **kwargs,
    )


AbstractFileSystem.read_json_file = read_json_file
AbstractFileSystem.read_json = read_json
AbstractFileSystem.read_csv_file = read_csv_file
AbstractFileSystem.read_csv = read_csv
AbstractFileSystem.read_parquet_file = read_parquet_file
AbstractFileSystem.read_parquet = read_parquet
AbstractFileSystem.read_files = read_files
AbstractFileSystem.pyarrow_dataset = pyarrow_dataset
AbstractFileSystem.pydala_dataset = pydala_dataset
AbstractFileSystem.pyarrow_parquet_dataset = pyarrow_parquet_dataset
AbstractFileSystem.write_parquet = write_parquet
AbstractFileSystem.write_json = write_json
AbstractFileSystem.write_csv = write_csv
AbstractFileSystem.write_file = write_file
AbstractFileSystem.write_files = write_files
AbstractFileSystem.write_pyarrow_dataset = write_pyarrow_dataset
AbstractFileSystem.write_pydala_dataset = write_pydala_dataset



import datetime as dt
import os
import posixpath
from pathlib import Path

import rich
from fsspec.spec import AbstractFileSystem

from .cfg import Config
from .fs import get_filesystem


def init(
    name: str | None = None,
    base_dir: str | None = None,
    storage_options: dict = {},
    fs: AbstractFileSystem | None = None,
):
    if name is None:
        name = str(Path.cwd().name)
        base_dir = str(Path.cwd().parent)

    if base_dir is None:
        base_dir = str(Path.cwd())

    fs = get_filesystem(posixpath.join(base_dir, name), **storage_options)

    fs.makedirs("conf/pipelines", exist_ok=True)
    fs.makedirs("pipelines", exist_ok=True)

    cfg = Config.load(base_dir=posixpath.join(base_dir, name), name=name)

    with open(posixpath.join(base_dir, name, "README.md"), "w") as f:
        f.write(
            f"# {name.replace('_', ' ').upper()}\n\n"
            f"**created with FlowerPower**\n\n*{dt.datetime.now().strftime('%Y-%m-%d %H:%M:%S')}*\n\n"
        )
    cfg.save()
    os.chdir(posixpath.join(base_dir, name))

    rich.print(
        f"\n✨ Initialized FlowerPower project [bold blue]{name}[/bold blue] "
        f"at [italic green]{base_dir}[/italic green]\n"
    )

    rich.print(
        """[yellow]Getting Started:[/yellow]

    📦  It is recommended to use the project manager [bold cyan]`uv`[/bold cyan] to manage the
        dependenvies of your project.

    🔧  Install uv:
            [dim]Run:[/dim] [bold white]pip install uv[/bold white]
            [dim]More options:[/dim]
                [blue underline]https://docs.astral.sh/uv/getting-started/installation/[/blue underline]

    🚀  Initialize your project:
            [dim]Run the following in your project directory:[/dim]
            [bold white]uv init --app --no-readme --vcs git[/bold white]
    """
    )


# def find_pipelines(cls):
#     """Find all pipeline modules in the project's pipelines directory."""
#     pipeline_path = Path("pipelines")
#     if not pipeline_path.exists():
#         return []

#     pipelines = []
#     for file in pipeline_path.glob("*.py"):
#         if file.name.startswith("_"):
#             continue

#         module_name = file.stem
#         try:
#             pipeline = Pipeline(module_name)
#             pipelines.append(pipeline)
#         except Exception as e:
#             rich.print(f"[red]Error loading pipeline {module_name}: {str(e)}[/red]")

#     return pipelines


# def list_pipelines():
#     pipelines = Pipeline.find_pipelines()
#     if not pipelines:
#         rich.print("\n📭 [yellow]No pipelines found in this project[/yellow]\n")
#         return

#     rich.print("\n🌸 [bold magenta]Available Pipelines:[/bold magenta]\n")
#     table = rich.table.Table(show_header=True, header_style="bold cyan")
#     table.add_column("Name")
#     table.add_column("Description")
#     table.add_column("Status")

#     for pipeline in pipelines:
#         status = (
#             "[green]Active[/green]" if pipeline.is_active() else "[red]Inactive[/red]"
#         )
#         table.add_row(
#             pipeline.name, pipeline.description or "[dim]No description[/dim]", status
#         )

#     rich.print(table)
#     rich.print()



from typing import Any

import datafusion
import duckdb
import orjson
import pandas as pd
import polars as pl
import pyarrow as pa
import pyarrow.dataset as pds
from pydantic import BaseModel, ConfigDict

from ...utils.sql import sql2polars_filter
from ..metadata import get_dataframe_metadata, get_duckdb_metadata


class PayloadReader(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    payload: bytes | dict[str, Any]
    topic: str | None = None
    conn: duckdb.DuckDBPyConnection | None = None
    ctx: datafusion.SessionContext | None = None
    format: str = "mqtt"

    def model_post_init(self, __context):
        if isinstance(self.payload, bytes):
            self.payload = orjson.loads(self.payload)

        self._metadata = {
            "format": self.format,
            "timestamp": pd.Timestamp.now(),
            "topic": self.topic,
        }

    def to_pyarrow_table(
        self, metadata: bool = False
    ) -> pa.Table | tuple[pa.Table, dict[str, Any]]:
        try:
            df = pa.Table.from_pydict(self.payload)
        except pa.ArrowInvalid:
            df = pa.Table.from_pylist([self.payload])
        if metadata:
            self._metadata = get_dataframe_metadata(df, **self._metadata)
            return df, self._metadata
        return df

    def to_pandas(
        self, metadata: bool = False
    ) -> pd.DataFrame | tuple[pd.DataFrame, dict[str, Any]]:
        try:
            df = pd.DataFrame(self.payload)
        except ValueError:
            df = pd.DataFrame([self.payload])
        if metadata:
            self._metadata = get_dataframe_metadata(df, **self._metadata)
            return df, self._metadata
        return df

    def _to_polars_dataframe(
        self, metadata: bool = False
    ) -> pl.DataFrame | tuple[pl.DataFrame, dict[str, Any]]:
        try:
            df = pl.DataFrame(self.payload)
        except pl.exceptions.ShapeError:
            df = pl.DataFrame([self.payload])
        if metadata:
            self._metadata = get_dataframe_metadata(df, **self._metadata)
            return df, self._metadata
        return df

    def _to_polars_lazyframe(
        self, metadata: bool = False
    ) -> pl.LazyFrame | tuple[pl.LazyFrame, dict[str, Any]]:
        try:
            df = pl.LazyFrame(self.payload)
        except pl.exceptions.ShapeError:
            df = pl.LazyFrame([self.payload])
        if metadata:
            self._metadata = get_dataframe_metadata(df, **self._metadata)
            return df, self._metadata
        return df

    def to_polars(
        self, lazy: bool = False, metadata: bool = False
    ) -> (
        pl.DataFrame | pl.LazyFrame | tuple[pl.DataFrame | pl.LazyFrame, dict[str, Any]]
    ):
        if lazy:
            return self._to_polars_lazyframe(metadata=metadata)
        else:
            return self._to_polars_dataframe(metadata=metadata)

    def to_duckdb_relation(
        self, conn: duckdb.DuckDBPyConnection | None = None, metadata: bool = False
    ) -> duckdb.DuckDBPyRelation | tuple[duckdb.DuckDBPyRelation, dict[str, Any]]:
        if self.conn is None:
            if conn is None:
                conn = duckdb.connect()
            self.conn = conn
        rel = self.conn.from_arrow(self.to_pyarrow_table())
        if metadata:
            self._metadata = get_duckdb_metadata(rel, **self._metadata)
            return rel, self._metadata
        return rel

    def to_pyarrow_dataset(
        self, metadata: bool = False, **kwargs
    ) -> pds.Dataset | tuple[pds.Dataset, dict[str, Any]]:
        if metadata:
            t, self._metadata = self.to_pyarrow_table(metadata=True)
            return pds.dataset(t, **kwargs), self._metadata
        return pds.dataset(self.to_pyarrow_table(), **kwargs)

    def register_in_duckdb(
        self,
        conn: duckdb.DuckDBPyConnection | None = None,
        name: str | None = None,
    ) -> duckdb.DuckDBPyConnection:
        if name is None:
            name = f"mqtt:{self.topic}"

        if self.conn is None:
            if conn is None:
                conn = duckdb.connect()
            self.conn = conn

        self.conn.register(name, self.to_pyarrow_table())
        return self.conn

    def register_in_datafusion(
        self,
        ctx: datafusion.SessionContext | None = None,
        name: str | None = None,
    ) -> None:
        if name is None:
            name = f"mqtt:{self.topic}"

        if self.ctx is None:
            if ctx is None:
                ctx = datafusion.SessionContext()
            self.ctx = ctx

        self.ctx.register(name, [self.to_pyarrow_table()])

        return self.ctx

    def filter(self, filter_expr: str | pl.Expr) -> pl.DataFrame | pl.LazyFrame:
        self._data = self.to_polars()

        pl_schema = (
            self._data.schema
            if isinstance(self._data, pl.DataFrame)
            else self._data.collect_schema()
        )
        filter_expr = (
            sql2polars_filter(filter_expr, pl_schema)
            if isinstance(filter_expr, str)
            else filter_expr
        )
        return self._data.filter(filter_expr)



from ..base import BaseDatabaseWriter


class MSSQLWriter(BaseDatabaseWriter):
    """MSSQL writer.

    This class is responsible for writing dataframes to MsSQL database.

    Examples:
        ```python
        writer = MSSQLWriter(table_name="table", host="localhost",
                                port=5432, username="user", password="password",
                                database="database")
        writer.write(df)

        # or
        writer = MSSQLWriter(table_name="table",
                                connection_string="mssql+pyodbc://user:password@localhost:5432/database")
        writer.write(df)
        ```
    """

    type_: str = "mssql"

    def model_post_init(self, __context):
        super().model_post_init(__context)



import importlib

has_orjson = importlib.util.find_spec("orjson") is not None

if has_orjson:
    from .ext import AbstractFileSystem
else:
    from fsspec import AbstractFileSystem

from .base import get_filesystem



import typer

from ..scheduler import SchedulerManager  # Adjust import as needed
from .utils import \
    parse_dict_or_list_param  # Assuming you have this utility function

# Create a Typer app for scheduler commands
app = typer.Typer(help="Scheduler management commands")


@app.command()
def start_worker(
    name: str | None = None,
    base_dir: str | None = None,
    background: bool = False,
    storage_options: str | None = None,
):
    """
    Start a scheduler worker.

    Args:
        name: Name of the scheduler
        base_dir: Base directory for the scheduler
        background: Run in background
        storage_options: Storage options as JSON or key=value pairs
    """
    parsed_storage_options = parse_dict_or_list_param(storage_options, "dict") or {}

    with SchedulerManager(
        name, base_dir, storage_options=parsed_storage_options
    ) as manager:
        manager.start_worker(background=background)


@app.command()
def remove_all_schedules(
    name: str | None = None,
    base_dir: str | None = None,
    storage_options: str | None = None,
):
    """
    Remove all schedules from the scheduler.

    Args:
        name: Name of the scheduler
        base_dir: Base directory for the scheduler
        storage_options: Storage options as JSON or key=value pairs
    """
    parsed_storage_options = parse_dict_or_list_param(storage_options, "dict") or {}

    with SchedulerManager(
        name, base_dir, role="scheduler", storage_options=parsed_storage_options
    ) as manager:
        manager.remove_all_schedules()


# @app.command()
# def add_schedule(
#     name: str | None = None,
#     base_dir: str | None = None,
#     storage_options: str | None = None,
#     trigger_type: str = "cron",
#     **kwargs,
# ) -> str:
#     """
#     Add a schedule to the scheduler.

#     Args:
#         name: Name of the scheduler
#         base_dir: Base directory for the scheduler
#         storage_options: Storage options as JSON or key=value pairs
#         trigger_type: Type of trigger (cron, interval, etc.)
#         **kwargs: Additional schedule parameters
#     """
#     parsed_storage_options = parse_dict_or_list_param(storage_options, "dict") or {}

#     with SchedulerManager(name, base_dir, role="scheduler") as manager:
#         schedule_id = manager.add_schedule(
#             storage_options=parsed_storage_options, type=trigger_type, **kwargs
#         )

#     typer.echo(f"Schedule added with ID: {schedule_id}")
#     return schedule_id


# @app.command()
# def add_job(
#     name: str | None = None,
#     base_dir: str | None = None,
#     storage_options: str | None = None,
#     **kwargs,
# ) -> str:
#     """
#     Add a job to the scheduler.

#     Args:
#         name: Name of the scheduler
#         base_dir: Base directory for the scheduler
#         storage_options: Storage options as JSON or key=value pairs
#         **kwargs: Job parameters
#     """
#     parsed_storage_options = parse_dict_or_list_param(storage_options, "dict") or {}

#     with SchedulerManager(name, base_dir, role="scheduler") as manager:
#         job_id = manager.add_job(storage_options=parsed_storage_options, **kwargs)

#     typer.echo(f"Job added with ID: {job_id}")
#     return str(job_id)


# @app.command()
# def run_job(
#     name: str | None = None,
#     base_dir: str | None = None,
#     storage_options: str | None = None,
#     **kwargs,
# ):
#     """
#     Run a job and return its result.

#     Args:
#         name: Name of the scheduler
#         base_dir: Base directory for the scheduler
#         storage_options: Storage options as JSON or key=value pairs
#         **kwargs: Job parameters
#     """
#     parsed_storage_options = parse_dict_or_list_param(storage_options, "dict") or {}

#     with SchedulerManager(name, base_dir, role="scheduler") as manager:
#         result = manager.run_job(storage_options=parsed_storage_options, **kwargs)

#     typer.echo("Job result:")
#     typer.echo(result)
#     return result


@app.command()
def get_schedules(
    name: str | None = None,
    base_dir: str | None = None,
    storage_options: str | None = None,
):
    """
    Get all schedules from the scheduler.

    Args:
        name: Name of the scheduler
        base_dir: Base directory for the scheduler
        storage_options: Storage options as JSON or key=value pairs
    """
    parsed_storage_options = parse_dict_or_list_param(storage_options, "dict") or {}

    with SchedulerManager(
        name, base_dir, role="scheduler", storage_options=parsed_storage_options
    ) as manager:
        schedules = manager.get_schedules()

    typer.echo("Schedules:")
    for schedule in schedules:
        typer.echo(schedule)
    return schedules


@app.command()
def get_tasks(
    name: str | None = None,
    base_dir: str | None = None,
    storage_options: str | None = None,
):
    """
    Get all tasks from the scheduler.

    Args:
        name: Name of the scheduler
        base_dir: Base directory for the scheduler
        storage_options: Storage options as JSON or key=value pairs
    """
    parsed_storage_options = parse_dict_or_list_param(storage_options, "dict") or {}

    with SchedulerManager(
        name, base_dir, role="scheduler", storage_options=parsed_storage_options
    ) as manager:
        tasks = manager.get_tasks()

    typer.echo("Tasks:")
    for task in tasks:
        typer.echo(task)
    return tasks


@app.command()
def get_jobs(
    name: str | None = None,
    base_dir: str | None = None,
    storage_options: str | None = None,
):
    """
    Get all jobs from the scheduler.

    Args:
        name: Name of the scheduler
        base_dir: Base directory for the scheduler
        storage_options: Storage options as JSON or key=value pairs
    """
    parsed_storage_options = parse_dict_or_list_param(storage_options, "dict") or {}

    with SchedulerManager(
        name, base_dir, role="scheduler", storage_options=parsed_storage_options
    ) as manager:
        jobs = manager.get_jobs()

    typer.echo("Jobs:")
    for job in jobs:
        typer.echo(job)
    return jobs


@app.command()
def get_job_result(
    job_id: str,
    name: str | None = None,
    base_dir: str | None = None,
    storage_options: str | None = None,
):
    """
    Get the result of a specific job.

    Args:
        job_id: ID of the job
        name: Name of the scheduler
        base_dir: Base directory for the scheduler
        storage_options: Storage options as JSON or key=value pairs
    """
    parsed_storage_options = parse_dict_or_list_param(storage_options, "dict") or {}

    with SchedulerManager(
        name, base_dir, role="scheduler", storage_options=parsed_storage_options
    ) as manager:
        result = manager.get_job_result(job_id)

    typer.echo("Job Result:")
    typer.echo(result)
    return result


@app.command()
def show_schedules(
    name: str | None = None,
    base_dir: str | None = None,
    storage_options: str | None = None,
):
    """
    Show all schedules in the scheduler.

    Args:
        name: Name of the scheduler
        base_dir: Base directory for the scheduler
        storage_options: Storage options as JSON or key=value pairs
    """
    parsed_storage_options = parse_dict_or_list_param(storage_options, "dict") or {}

    with SchedulerManager(
        name, base_dir, role="scheduler", storage_options=parsed_storage_options
    ) as manager:
        manager.show_schedules()


@app.command()
def show_jobs(
    name: str | None = None,
    base_dir: str | None = None,
    storage_options: str | None = None,
):
    """
    Show all jobs in the scheduler.

    Args:
        name: Name of the scheduler
        base_dir: Base directory for the scheduler
        storage_options: Storage options as JSON or key=value pairs
    """
    parsed_storage_options = parse_dict_or_list_param(storage_options, "dict") or {}

    with SchedulerManager(
        name, base_dir, role="scheduler", storage_options=parsed_storage_options
    ) as manager:
        manager.show_jobs()


@app.command()
def show_tasks(
    name: str | None = None,
    base_dir: str | None = None,
    storage_options: str | None = None,
):
    """
    Show all tasks in the scheduler.

    Args:
        name: Name of the scheduler
        base_dir: Base directory for the scheduler
        storage_options: Storage options as JSON or key=value pairs
    """
    parsed_storage_options = parse_dict_or_list_param(storage_options, "dict") or {}

    with SchedulerManager(
        name, base_dir, role="scheduler", storage_options=parsed_storage_options
    ) as manager:
        manager.show_tasks()



from ..base import BaseDatabaseReader


class DuckDBReader(BaseDatabaseReader):
    """DuckDB loader.

    This class is responsible for loading dataframes from DuckDB database.

    Examples:
        ```python
        loader = DuckDBReader(table_name="table", path="data.db")
        df = loader.to_polars("SELECT * FROM table WHERE column = 'value'")
        ```
    """

    type_: str = "duckdb"

    def model_post_init(self, __context):
        super().model_post_init(__context)



# PIPELINE_TEMPLATE = """# ---------------- Pipelines Configuration ----------------- #

# # ------------------------ Example ------------------------- #
# #
# # path: pipelines
# #
# # ## pipeline parameter
# #
# # params:
# #   flow1:                      ## pipeline name
# #       step1:                  ## step name
# #         param1_1: 123         ## step parameters
# #         param1_2: abc
# #       step2:
# #         param2_1: true
# #
# # ## run configuration
# #
# # run:
# #   prod: # environment name
# #     flow1:
# #       inputs:                 ## input parameters
# #       final_vars: [step2]     ## final output vars
# #       with_tracker: true      ## whether to track the run
# #
# #   dev:
# #     flow1:
# #       inputs:
# #       final_vars: [step2]
# #       with_tracker: false
# #
# # ---------------------------------------------------------- #

# """

# SCHEDULER_TEMPLATE = """# ---------------- Scheduler Configuration ----------------- #

# # ------------------------ Example ------------------------- #
# #
# # ##  data store configuration
# #
# # ### postgres
# # data_store:
# #   type: sqlalchemy
# #   url: postgresql+asyncpg://edge:edge@postgres/flowerpower
# #
# # ### sqlite
# # data_store:
# #   type: sqlalchemy
# #   url: sqlite+aiosqlite:///flowerpower.db
# #
# # ### memory
# # data_store:
# #   type: memory
# #
# # ### mongodb
# # data_store:
# #   type: mongodb
# #   url: mongodb://localhost:27017/scheduler
# #
# # ## event broker configuration
# #
# # ### postgres
# # event_broker:
# #   type: asyncpg
# #   url: postgresql+asyncpg://edge:edge@postgres/flowerpower
# #
# # ### mqtt
# # event_broker:
# #   type: mqtt
# #   host: localhost
# #   port: 1883
# #   username: edge
# #   password: edge

# # ### redis
# # event_broker:
# #   type: redis
# #   host: localhost
# #   port: 6379

# # ### local
# # event_broker:
# #   type: local # or memory
# #
# # ## cleanup interval configuration
# #
# # cleanup_interval:
# #   unit: minutes
# #   value: 15
# #
# # ## pipeline schedule configuration
# #
# # pipeline:
# #   my_flow:
# #     type: cron                ## options: interval, calendarinterval, date
# #     start_time:
# #     end_time:
# #     ## optional cron arguments
# #     crontab: * * * * *
# #     year:
# #     month:
# #     week:
# #     day:
# #     days_of_week:
# #     hour:
# #     minute:
# #     second:
# #     timezone:
# #     ## optional interval arguments
# #     weeks:
# #     days:
# #     hours:
# #     minutes:
# #     seconds:
# #     microseconds:
# #
# # ---------------------------------------------------------- #

# """

# TRACKER_TEMPLATE = """# ----------------- Tracker Configuration ------------------ #

# # ------------------------ Example ------------------------- #
# #
# # username: your.email@example.com
# # api_url: http://localhost:8241
# # ui_url: http://localhost:8242
# # api_key:

# # pipeline:
# #   my_flow:
# #     project_id: 1
# #     tags:
# #       environment: dev
# #       version: 1.0
# #       TODO: add_more_tags_to_find_your_run_later
# #     dag_name: my_flow_123
# #
# # ---------------------------------------------------------- #

# """

PIPELINE_PY_TEMPLATE = """# FlowerPower pipeline {name}.py
# Created on {date}

####################################################################################################
# Import necessary libraries
# NOTE: Remove or comment out imports that are not used in the pipeline

from hamilton.function_modifiers import parameterize, dataloader, datasaver
from hamilton.htypes import Parallelizable, Collect

from pathlib import Path

from flowerpower.cfg import Config

####################################################################################################
# Load pipeline parameters. Do not modify this section.

PARAMS = Config.load(
    Path(__file__).parents[1], pipeline_name="{name}"
).pipeline.h_params


####################################################################################################
# Helper functions.
# This functions have to start with an underscore (_).


####################################################################################################
# Pipeline functions

"""



from ..base import BaseDatasetReader, BaseFileReader


class ParquetFileReader(BaseFileReader):
    """Parquet file loader.

    This class is responsible for loading dataframes from Parquet files.

    Examples:
        ```python
        loader = ParquetFileReader("data.parquet")
        df = loader.load()
        ```
    """

    format: str = "parquet"

    def model_post_init(self, __context):
        super().model_post_init(__context)


class ParquetDatasetReader(BaseDatasetReader):
    """Parquet dataset loader.

    This class is responsible for loading dataframes from Parquet dataset.

    Examples:
        ```python
        loader = ParquetDatasetReader("parquet_data/")
        df = loader.load()
        ```
    """

    format: str = "parquet"

    def model_post_init(self, __context):
        super().model_post_init(__context)
